# imports
import matplotlib.pyplot as plt
from getdist import plots
import sys
import os


# corner plot
def corner(mcsamples, markers, output=None, noshow=False, filled_alpha=None, contour_alpha=0.5):
    colors = ["#006FED", "#E03424", "#008000", "#9c5500", "#9224e0", "#ed00e6", "#f2e400", "#00f2e4", "#6fd95f"]
    contour_colors = ["#006FED", "#E03424", "#008000", "#9c5500", "#9224e0", "#ed00e6", "#f2e400", "#00f2e4", "#6fd95f"]

    g = plots.get_subplot_plotter()
    g.settings.alpha_factor_contour_lines=contour_alpha

    if filled_alpha:
        alpha = []
        for i in range(len(mcsamples)):
            alpha.append({"alpha": filled_alpha})
        g.triangle_plot(mcsamples, filled=True, markers=markers, colors=colors, contour_colors=contour_colors, contour_args=alpha)
    else:
        g.triangle_plot(mcsamples, filled=True, markers=markers, contour_colors=contour_colors, colors=colors)

    if output:
        plt.savefig(output, transparent=True)
    if not noshow:
        plt.show()
    plt.close()

    return


# print system information
def syslog(file=sys.stdout):
    if file != sys.stdout:
        file = open(file, "w")
        print("## sys.log", file=file)
        print("# information regarding the system and the date in which this run was executed", file=file)
        print("", file=file)

    print("$ date", file=file)
    date = os.popen("date").read()[:-1]
    print(f"{date}", file=file)
    print("", file=file)

    print("$ uname -a", file=file)
    uname = os.popen("uname -a").read()[:-1]
    print(f"{uname}", file=file)
    print("", file=file)

    print("$ lscpu", file=file)
    lscpu = os.popen("lscpu").read()[:-1]
    print(f"{lscpu}", file=file)
    print("", file=file)

    if file != sys.stdout:
        file.close()

    return


# print confidence intervals in a latex table
def CIs(mcsamples, file=sys.stdout):
    if file != sys.stdout:
        file = open(file, "w")

    print("## CIs.tex", file=file)
    print("# latex table for the 1 and 2 sigma distribution of each parameter", file=file)
    print("", file=file)

    print(mcsamples.getTable(limit=1).tableTex().replace("\n\n", "\n"), file=file)
    print("", file=file)
    print(mcsamples.getTable().tableTex().replace("\n\n", "\n"), file=file)

    if file != sys.stdout:
        file.close()

    return



# imports
from multiprocessing import cpu_count
from random import gauss, uniform
import matplotlib.pyplot as plt
import numpy as np
import yaml
import sys
import os


# fetch the arguments from CLI and from the configuration file and check for incompatible or missing arguments
def load(args):
    # required argumentss
    model = args.model
    data = args.data

    # config file
    if args.yml:
        yml = args.yml
    else:
        os.system("echo 'none: none' > /tmp/dummy.yml")
        yml = "/tmp/dummy.yml"

    # get config arguments from file or CLI
    with open(yml, "r") as file:
        yml_loaded = yaml.full_load(file)
        names = eval(args.names) if args.names else yml_loaded.get("names")
        labels = eval(args.labels) if args.labels else yml_loaded.get("labels")
        initial = eval(args.initial) if args.initial else yml_loaded.get("initial")
        markers = eval(args.markers) if args.markers else yml_loaded.get("markers")
        samples = args.samples if args.samples else yml_loaded.get("samples")
        warmup = args.warmup if args.warmup else yml_loaded.get("warmup")
        chains = args.chains if args.chains else yml_loaded.get("chains")

    # output arguments
    output = args.output
    savechain = args.save_chain
    gzip = args.gzip
    lzf = args.lzf
    noshow = args.no_show

    # check if everything is provided
    if not names:
        raise Exception("Parameters names must be provided either in CLI or configuration file")
    if not labels:
        labels = names
    if not initial:
        raise Exception("Initial confitions must be provided either in CLI or configuration file")
    if not samples:
        raise Exception("The number of steps to sample the posterior distribution, after the warmup, must be provided either in CLI or configuration file")
    if not warmup:
        raise Exception("The number of steps to warmup each chain must be provided either in CLI or configuration file")

    # set default values if not provided
    if not chains:
        chains = cpu_count()

    if not markers:
        markers = {}
    for name in names:
        try:
            markers[name]
        except KeyError:
            markers[name] = None

    if gzip == []:
        gzip = 4
    elif gzip:
        gzip = int(gzip[0])

    # check gzip values
    if gzip:
        if gzip < 0 or gzip > 9:
            raise Exception(f"Value of gzip must be between 0 and 9 (inclusive), provided value was {gzip}")

    # check if sizes match
    if not ( len(names) == len(labels) == len(initial) ):
        raise Exception(f"number of dimensions missmatch: len(names) = {len(names)}, len(labels) = {len(labels)}, len(initial) = {len(initial)}")

    # check for single compression algorithm
    if gzip and lzf:
        raise Exception("--gzip and --lzf are mutually exclusive, pick one compression algorithm")

    # if noshow is provided, output must also be provided
    if noshow and not output:
        raise Exception("Toggling -n, --noshow requires to provide an output folder, otherwise output will not be shown nor saved.")

    # evaluate initial conditions to Python function(s), for each chain
    # we're returning both init and initial because the latest is required to output the configuration used
    init = []
    for i in range(0, chains):
        init.append({})
        for name in names:
            init[i][name] = eval(initial[name])

    # number of parameters, useful later
    ndim = len(names)

    return model, data, yml, names, labels, initial, markers, samples, warmup, chains, output, savechain, gzip, lzf, noshow, init, ndim


# save configuration used to file
def save(yml, names, labels, initial, markers, samples, warmup, chains, outputyml, output, savechain, gzip, lzf, noshow):
    # orderly save the configuration options
    with open(yml, "w") as file:
        file.write("## config.yml\n")
        file.write("# backup of all the configuration arguments used for this specific run\n")
        file.write("\n")

        yaml.dump({"names": names}, file)
        file.write("\n")

        yaml.dump({"labels": labels}, file)
        file.write("\n")

        yaml.dump({"initial": initial}, file, sort_keys=False)
        file.write("\n")

        yaml.dump({"markers": markers}, file, sort_keys=False)
        file.write("\n")

        yaml.dump({"samples": samples}, file)
        file.write("\n")

        yaml.dump({"warmup": warmup}, file)
        file.write("\n")

        yaml.dump({"chains": chains}, file)

    # orderly save the output arguments
    with open(outputyml, "w") as file:
        file.write("## output.yml\n")
        file.write("# backup of all the output arguments used for this specific run\n")
        file.write("\n")

        yaml.dump({"output": output}, file)
        file.write("\n")

        yaml.dump({"save-chain": savechain}, file)
        file.write("\n")

        yaml.dump({"gzip": gzip}, file)
        file.write("\n")

        yaml.dump({"lzf": lzf}, file)
        file.write("\n")

        yaml.dump({"noshow": noshow}, file)
        file.write("\n")

    return


# convert fit to a numpy array of size [steps, chains, ndim], with all of the computed steps
def getsteps(fit, names, samples, warmup, chains, ndim):
    totalsteps = np.empty([samples+warmup, chains, ndim])
    for i in range(ndim):
        for j in range(chains):
            totalsteps[:, j, i] = fit[names[i]][0][j::chains]

    return totalsteps


# flatten total steps (i.e. remove chain information) and remove warmup to a numpy array of size [steps, ndim]
def getflatsamples(samples, warmup, chains, ndim, totalsteps):
    flatsamples = np.empty([samples*chains, ndim])
    for i in range(ndim):
        start = 0
        for j in range(chains):
            flatsamples[start::chains, i] = totalsteps[warmup:, j, i]
            start += 1

    return flatsamples


# plot time series
def timeseries(totalsteps, names, labels, markers, samples, warmup, chains, ndim, output=None, noshow=False):
    fig, axes = plt.subplots(ndim, figsize=(10, 7), sharex=True)
    steps = np.arange(samples+warmup)

    for i in range(ndim):
        ax = axes[i]
        for j in range(chains):
            ax.plot(steps, totalsteps[:, j, i], alpha=0.75)
        ax.set_xlim(0, samples+warmup)
        ax.set_ylabel("$" + labels[i] + "$")
        ax.axvline(x=warmup, linestyle="--", color="black", alpha=0.5)
        if markers[names[i]]:
            ax.axhline(y=markers[names[i]], linestyle="--", color="black", alpha=0.5)
        ax.yaxis.set_label_coords(-0.1, 0.5)
        ax.grid()

    axes[-1].set_xlabel("step number")
    if output:
        plt.savefig(output)
    if not noshow:
        plt.show()
    plt.close()

    return


# print run information
def runlog(timeelapsed, file=sys.stdout):
    if file != sys.stdout:
        file = open(file, "w")
        print("## run.log", file=file)
        print("# information regarding the execution of this program", file=file)
        print("# the execution time has the format hours:minutes:seconds", file=file)
        print("", file=file)

    print(f"time: {timeelapsed}", file=file)

    if file != sys.stdout:
        file.close()

    return



# imports
from multiprocessing import cpu_count
import matplotlib.patches as mpatches
from random import gauss, uniform
import matplotlib.pyplot as plt
import numpy as np
import yaml
import sys


# fetch the arguments from CLI and the configuration file
# check for incompatible or missing arguments
# set default arguments value
def load(args):
    # required arguments
    model = args.model
    data = args.data

    # config file
    if args.yml:
        yml = args.yml
    else:
        os.system("echo 'none: none' > /tmp/dummy.yml")
        yml = "/tmp/dummy.yml"

    # get config arguments from file and overwrite if provided in the CLI
    with open(yml, "r") as file:
        yml_loaded = yaml.full_load(file)

        names = eval(args.names) if args.names else yml_loaded.get("names")
        labels = eval(args.labels) if args.labels else yml_loaded.get("labels")
        initial = eval(args.initial) if args.initial else yml_loaded.get("initial")
        markers = eval(args.markers) if args.markers else yml_loaded.get("markers")
        percentage = args.percentage / 100 if args.percentage else yml_loaded.get("percentage") / 100
        samples = args.samples if args.samples else yml_loaded.get("samples")
        check = args.check if args.check else yml_loaded.get("check")
        maxsteps = args.maxsteps if args.maxsteps else yml_loaded.get("maxsteps")
        walkers = args.walkers if args.walkers else yml_loaded.get("walkers")
        processes = args.processes if args.processes else yml_loaded.get("processes")

    # output arguments
    output = args.output
    savechain = args.save_chain
    gzip = args.gzip
    lzf = args.lzf
    tmp = args.tmp
    shm = args.shm
    thin = args.thin
    timeseries = args.time_series
    noshow = args.no_show
    noprogress = args.no_progress

    # check if everything that is required is provided
    if not names:
        raise Exception("Parameters names must either be provided in CLI or in the configuration file")
    if not labels:
        labels = names
    if not initial:
        raise Exception("Initial confitions must either be provided in CLI or in the configuration file")
    if not percentage:
        raise Exception("The percentage to consider that convergence is met must either be provided in the CLI or in the configuration file")
    if not samples:
        raise Exception("The number of samples to compute when converge is met must either be provided in CLI or in the configuration file")

    # set defaults
    if not check:
        check = 1000

    if not maxsteps:
        maxsteps = 100000

    if not walkers:
        walkers = 32

    if not processes:
        processes = cpu_count()

    if not markers:
        markers = {}
    for name in names:
        try:
            markers[name]
        except KeyError:
            markers[name] = None

    # check if sizes match
    if not ( len(names) == len(labels) == len(initial) ):
        raise Exception(f"number of dimensions missmatch: len(names) = {len(names)}, len(labels) = {len(labels)}, len(initial) = {len(initial)}")

    # exit if output is not provided and noshow is
    if not output and noshow:
        raise Exception("Flag -n, --noshow is provided without providing -o, --output. This means that the output will not be shown nor saved")

    # exit if savechain is provided but output is not
    if not output and savechain:
        raise Exception("Flag --save-chain is set, but output wasn't provided (using --output or -o)")

    # exit if both tmp and shm are provided
    if tmp and shm:
        raise Exception("Flags --tmp and --shm are mutually exclusive, pick the one which is mounted as tmpfs in your system")

    # exit if tmp or shm is provided, but chain is not
    if (tmp or shm) and not savechain:
        raise Exception("Flag --tmp requires the usage of --save-chain")

    # exit if the number of steps to compute is lower than the maximum number of steps
    if maxsteps < samples:
        raise Exception("The maximum number of steps (-M, --maxsteps) must always be larger than the number of steps to compute (-s, --samples)")

    # check for single compression algorithm
    if gzip and lzf:
        raise Exception("--gzip and --lzf are mutually exclusive, pick one compression algorithm")

    # evaluate initial conditions to Python functions, and turn into an emcee compatible numpy array
    # we're returning both init and initial because the latest is required to output the configuration used
    init = np.empty([walkers, len(names)])
    for i in range(0, walkers):
        for j in range(0, len(names)):
            init[i][j] = eval(initial[names[j]])

    # auxiliary varialbel required here and there
    ndim = len(names)

    return model, data, yml, names, labels, initial, markers, percentage, samples, check, maxsteps, walkers, processes, output, savechain, gzip, lzf, tmp, shm, thin, timeseries, noshow, noprogress, init, ndim


# save configuration and output arguments to files
def save(yml, names, labels, initial, markers, percentage, samples, check, maxsteps, walkers, processes, outputyml, output, savechain, gzip, lzf, tmp, shm, thin, timeseries, noshow, noprogress):
    # orderly save the configuration arguments
    with open(yml, "w") as file:
        file.write("## config.yml\n")
        file.write("# backup of all the configuration arguments used for this specific run\n")
        file.write("\n")

        yaml.dump({"names": names}, file)
        file.write("\n")

        yaml.dump({"labels": labels}, file)
        file.write("\n")

        yaml.dump({"initial": initial}, file, sort_keys=False)
        file.write("\n")

        yaml.dump({"markers": markers}, file, sort_keys=False)
        file.write("\n")

        yaml.dump({"percentage": percentage * 100}, file)
        file.write("\n")

        yaml.dump({"samples": samples}, file)
        file.write("\n")

        yaml.dump({"check": check}, file)
        file.write("\n")

        yaml.dump({"maxsteps": maxsteps}, file)
        file.write("\n")

        yaml.dump({"walkers": walkers}, file)
        file.write("\n")

        yaml.dump({"processes": processes}, file)
        file.write("\n")

    # orderly save the output arguments
    with open(outputyml, "w") as file:
        file.write("## output.yml\n")
        file.write("# backup of all the output arguments used for this specific run\n")
        file.write("\n")

        yaml.dump({"output": output}, file)
        file.write("\n")

        yaml.dump({"save-chain": savechain}, file)
        file.write("\n")

        yaml.dump({"gzip": gzip}, file)
        file.write("\n")

        yaml.dump({"lzf": lzf}, file)
        file.write("\n")

        yaml.dump({"tmp": tmp}, file)
        file.write("\n")

        yaml.dump({"shm": shm}, file)
        file.write("\n")

        yaml.dump({"thin": thin}, file)
        file.write("\n")

        yaml.dump({"timeseries": timeseries}, file)
        file.write("\n")

        yaml.dump({"noshow": noshow}, file)
        file.write("\n")

        yaml.dump({"noprogress": noprogress}, file)
        file.write("\n")

    return

def autocorrelation(correlation, samples, check, index, laststep, delta, output=None, noshow=False):
    if type(correlation) != tuple:
        correlation = (correlation,)

    for autocorr in correlation:
        n = check * np.arange(1, index + 1)
        y = autocorr[:index]
        xmin = 0
        xmax = n.max() + check
        ymin = y.min() - 0.1 * (y.max() - y.min())
        ymax = y.max() + 0.1 * (y.max() - y.min())

        plt.plot(n, y, marker=".")

        region = mpatches.Rectangle((laststep-samples, autocorr[index-1] - delta), samples, 2*delta, color="red", alpha=0.2, label="convergence region")
        plt.gca().add_patch(region)

    plt.grid()
    plt.xlim(xmin, xmax)
    plt.ylim(ymin, ymax)
    plt.legend(handles=[region])
    plt.xlabel("number of steps")
    plt.ylabel(r"mean $\hat{\tau}$")
    if output:
        plt.savefig(output)
    if not noshow:
        plt.show()
    plt.close()

    return


# plot the time series
def timeseries(steps, labels, ndim, discard, output=None, noshow=False):
    fig, axes = plt.subplots(ndim, figsize=(10, 7), sharex=True)

    for i in range(ndim):
        ax = axes[i]
        ax.plot(steps[:, :, i], "k", alpha=0.3)
        ax.set_xlim(0, len(steps))
        ax.set_ylabel("$" + labels[i] + "$")
        ax.axvline(x=discard, linestyle="--", color="red")
        ax.yaxis.set_label_coords(-0.1, 0.5)

    axes[-1].set_xlabel("step number")
    if output:
        plt.savefig(output)
    if not noshow:
        plt.show()
    plt.close()

    return


# print run information
def runlog(timeelapsed, samples, discard, converged, file=sys.stdout):
    if file != sys.stdout:
        file = open(file, "w")

    print("## run.log", file=file)
    print("# information regarding the execution of this program", file=file)
    print("# the execution time has the format hours:minutes:seconds", file=file)
    print("", file=file)

    print(f"time: {timeelapsed}", file=file)
    print(f"converged: {converged > samples}", file=file)
    print(f"samples: {samples}", file=file)
    print(f"discard: {discard}", file=file)

    if file != sys.stdout:
        file.close()

    return



from setuptools import setup

with open(f"README.md") as f:
    long_description = f.read()

setup(name="simplifiedmc",
      version="1.0.1",
      description="A CLI that simplifies the usage of MCMC methods.",
      scripts=["bin/smc-stan", "bin/smc-emcee", "bin/smc-analyze"],
      packages=["simplifiedmc"],
      long_description=long_description,
      long_description_content_type="text/markdown",
      install_requires=[
        "numpy",
        "matplotlib",
        "emcee",
        "getdist",
        "tqdm",
        "h5py",
        "arviz",
        "pandas",
        "pystan",
        "pyyaml",
      ],
      url="https://github.com/jpmvferreira/simplifiedmc",
      author="José Ferreira",
      author_email="jose@jpferreira.me",
      license="MIT",
      zip_safe=False)
